
/*
  WiFi connected Woord LED Clock. It gets NTP time from the internet and translates to two CJMCU 64 Bit WS2812 5050 RGB LED Driver Development Board

  matrix arangement
  0 - 7   64 - 71
  8       72
  16      80
  24      88
  32      96
  40      104
  48      112
  56 - 63 120 - 127

  version <3
  16x8
  0123456701234567
  Het is vijf tien  0-2 4-5 7+64-66 68-71
  kwart  voor over  8-12 15+72-74 76-79
  half tween drie   16-19 21-23+80 23+80-81 83-86
  viervijfzeszeven  24-27 28-31 88-90 91-95
  achtnegentienelf  32-35 36-39+96 97-100 101-103
  twaalf uur in de  40-45 47+104-105 107-108 110-111
       middagavond  53-55+112-114 115-119
  1234nachtochtend  56-59 60-63+120 121-127

  version 3
  16x8
  0123456701234567
  Het is vijf tien  0-2 4-5 7+64-66 68-71
  kwart  voor over  8-12 15+72-74 76-79
  half tweeeendrie  16-19 21-23+80 81-83 84-87
  viervijfzeszeven  24-27 28-31 88-90 91-95
  achtnegentienelf  32-35 36-39+96 97-100 101-103
  twaalf uur in de  40-45 47+104-105 107-108 110-111
       middagavond  53-55+112-114 115-119
  1234nachtochtend  56-59 60-63+120 121-127


  Time handling
  Derived from round LED clock from, Leon van den Beukel, march 2019

  NTP and summer time code based on:
  https://tttapa.github.io/ESP8266/Chap15%20-%20NTP.html
  https://github.com/SensorsIot/NTPtimeESP/blob/master/NTPtimeESP.cpp (for US summer time support check this link)

  color handling
  https://github.com/FastLED/FastLED/wiki/Overview

*/

#define versie 3

#include <ESP8266WiFi.h>
#include <ESP8266WiFiMulti.h>
#include <WiFiUdp.h>
#include <FastLED.h>
#define DEBUG_ON

FASTLED_USING_NAMESPACE

#include "Secrets.h"             // this file is intentionally not included in the sketch, so nobody accidentally commits their secret information.
// create a Secrets.h file in the same directory as your schets with the following:
// #define WIFI_SSID "your-ssid"
// #define WIFI_PASS "your-password"
const char ssid[] = WIFI_SSID;        // Your network SSID name here
const char pass[] = WIFI_PASS;    // Your network password here

unsigned long timeZone = 1.0;    // Change this value to your local timezone (in my case +1 for Amsterdam)

CRGBPalette16 currentScheme;
byte colorschemeCount = 13;                           // number of colorscheme defined in setColorscheme
byte selectedColorscheme = 0;                         // default index

#define colorFixed ColorFromPalette(currentScheme,0)    // 0 - 255
#define colorHour ColorFromPalette(currentScheme,85)
#define colorMinute ColorFromPalette(currentScheme,170)
#define colorDag ColorFromPalette(currentScheme,255)

// if you dont want to use pallets but fixed color:
//CRGB colorFixed = CRGB::FloralWhite;
//CRGB colorHour = CRGB::Green;  // Red
//CRGB colorMinute = CRGB::Yellow; //Green;
//CRGB colorDag = CRGB::Purple; //Blue;

ESP8266WiFiMulti wifiMulti;
WiFiUDP UDP;
IPAddress timeServerIP;
const char* NTPServerName = "time.nist.gov";
const int NTP_PACKET_SIZE = 48;
byte NTPBuffer[NTP_PACKET_SIZE];

unsigned long intervalNTP = 1 * 60000; // Request NTP time every 1 minutes
unsigned long prevNTP = 0;
unsigned long lastNTPResponse = millis();
uint32_t timeUNIX = 1638701389;   // just to make a start time in case no NTP responce. was 0
unsigned long prevActualTime = 0;

#define LEAP_YEAR(Y) ( ((1970+Y)>0) && !((1970+Y)%4) && ( ((1970+Y)%100) || !((1970+Y)%400) ) )
static const uint8_t monthDays[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

#define VOLTS          5
#define MAX_MA       500

#define MAX_BRIGHTNESS  70
#define MIN_BRIGHTNESS  10
#define BRIGHTNESS_STEP 5
byte brightness = 15;

#define LED_TYPE    WS2812
#define COLOR_ORDER GRB
#define NUM_LEDS 128
#define DATA_PIN 2 // D4 
CRGB LEDs[NUM_LEDS];

#define BUTTON_PIN 0 // D3 pull-up
bool Buttonpressed = false;
unsigned long Buttontimer = 0;
#define LONGPRESS 700   // >700 millis 

struct DateTime {
  int  year;
  byte month;
  byte day;
  byte hour;
  byte minute;
  byte second;
  byte dayofweek;
};

DateTime currentDateTime;

void setup() {

  FastLED.delay(3000);
  FastLED.setMaxPowerInVoltsAndMilliamps( VOLTS, MAX_MA);
  // FastLED.addleds
  FastLED.addLeds<LED_TYPE, DATA_PIN, COLOR_ORDER>(LEDs, NUM_LEDS).setCorrection(TypicalSMD5050);
  FastLED.setBrightness(brightness);
  // FastLED.addLeds<WS2812B, DATA_PIN, GRB>(LEDs, NUM_LEDS);

  Serial.begin(115200);
  delay(10);
  Serial.println("\r\n");

  //configure pin 0 as an input and enable the internal pull-up resistor
  pinMode(BUTTON_PIN, INPUT_PULLUP);

  startWiFi();
  startUDP();

  if (!WiFi.hostByName(NTPServerName, timeServerIP)) {
    Serial.println("DNS lookup failed. Rebooting.");
    Serial.flush();
    ESP.reset();
  }
  Serial.print("Time server IP:\t");
  Serial.println(timeServerIP);

  Serial.println("\r\nSending NTP request ...");
  sendNTPpacket(timeServerIP);

  setColorscheme();
}

void loop() {
  unsigned long currentMillis = millis();

  if (currentMillis - prevNTP > intervalNTP) { // If a minute has passed since last NTP request
    prevNTP = currentMillis;
    Serial.println("\r\nSending NTP request ...");
    sendNTPpacket(timeServerIP);               // Send an NTP request
  }

  uint32_t time = getTime();                   // Check if an NTP response has arrived and get the (UNIX) time
  if (time) {                                  // If a new timestamp has been received
    timeUNIX = time;
    Serial.print("NTP response:\t");
    Serial.println(timeUNIX);
    lastNTPResponse = currentMillis;
  } else if ((currentMillis - lastNTPResponse) > 3600000) {
    Serial.println("More than 1 hour since last NTP response. Rebooting.");
    Serial.flush();
    ESP.reset();
  }

  int Button = digitalRead(BUTTON_PIN);
  if (Button == LOW && !Buttonpressed) {    // it means it is pressed
    Buttonpressed = true;
    Buttontimer = currentMillis;
  }
  if (Button == LOW && Buttonpressed) {
    if (currentMillis - Buttontimer >= LONGPRESS) {
      // select other colorscheme or pallet
      selectedColorscheme++;
      if (selectedColorscheme >= colorschemeCount) selectedColorscheme = 0;
      setColorscheme();
      for (int i = 0; i < NUM_LEDS; i++)
        LEDs[i] = CRGB::Black;
      prepareLEDtime();
      FastLED.show();
      delay(1000);
    }
  }
  if (Button == HIGH && Buttonpressed) {    // it is released
    Buttonpressed = false;
    if (currentMillis - Buttontimer < LONGPRESS) {   // it was a short press
      // increase brigtness
      brightness = brightness + BRIGHTNESS_STEP;
      if (brightness > MAX_BRIGHTNESS) brightness = MIN_BRIGHTNESS;
      FastLED.setBrightness(brightness);
      FastLED.show();
    }
  }


  uint32_t actualTime = timeUNIX + (currentMillis - lastNTPResponse) / 1000;
  if (actualTime != prevActualTime && timeUNIX != 0) { // If a second has passed since last update

    prevActualTime = actualTime;
    convertTime(actualTime);

    for (int i = 0; i < NUM_LEDS; i++)
      LEDs[i] = CRGB::Black;

    prepareLEDtime();

    //LEDs[getLEDMinuteOrSecond(currentDateTime.second)] = colorSecond;
    //LEDs[getLEDHour(currentDateTime.hour)] = colorHour;
    //LEDs[getLEDMinuteOrSecond(currentDateTime.minute)] = colorMinute;

    FastLED.show();
  }
}

void prepareLEDtime() {
  // Het is

  LEDs[0] = colorFixed;
  LEDs[1] = colorFixed;
  LEDs[2] = colorFixed;
  LEDs[4] = colorFixed;
  LEDs[5] = colorFixed;

  // minutes
  byte minutes = currentDateTime.minute;
  if (minutes >= 0 && minutes <= 4) {
    // heel uur
    LEDs[47] = colorHour;
    LEDs[104] = colorHour;
    LEDs[105] = colorHour;
    fillminutedetail(0 + minutes);
  } else if (minutes >= 5 && minutes <= 9) {
    // vijf
    fillvijf();
    // over
    fillover();
    fillminutedetail(minutes - 5);
  } else if (minutes >= 10 && minutes <= 14) {
    // tien
    filltien();
    // over
    fillover();
    fillminutedetail(minutes - 10);
  } else if (minutes >= 15 && minutes <= 19) {
    // kwart
    fillkwart();
    // over
    fillover();
    fillminutedetail(minutes - 15);
  } else if (minutes >= 20 && minutes <= 24) {
    // tien
    filltien();
    // voor
    fillvoor();
    fillhalf();
    fillminutedetail(minutes - 20);
  } else if (minutes >= 25 && minutes <= 29) {
    fillvijf();
    fillvoor();
    fillhalf();
    fillminutedetail(minutes - 25);
  } else if (minutes >= 30 && minutes <= 34) {
    // half
    fillhalf();
    fillminutedetail(minutes - 30);
  } else if (minutes >= 35 && minutes <= 39) {
    fillvijf();
    fillover();
    fillhalf();
    fillminutedetail(minutes - 35);
  } else if (minutes >= 40 && minutes <= 44) {
    filltien();
    fillover();
    fillhalf();
    fillminutedetail(minutes - 40);
  } else if (minutes >= 45 && minutes <= 49) {
    fillkwart();
    fillvoor();
    fillminutedetail(minutes - 45);
  } else if (minutes >= 50 && minutes <= 54) {
    filltien();
    fillvoor();
    fillminutedetail(minutes - 50);
  } else if (minutes >= 55 && minutes <= 59) {
    fillvijf();
    fillvoor();
    fillminutedetail(minutes - 55);
  }
  // hour
  byte hours = currentDateTime.hour;
  // dagdeel
  if (hours >= 0 && hours <= 5) {
    // nacht
    LEDs[60] = colorDag;
    LEDs[61] = colorDag;
    LEDs[62] = colorDag;
    LEDs[63] = colorDag;
    LEDs[120] = colorDag;
  } else if (hours >= 6 && hours <= 11) {
    // ochtend
    LEDs[121] = colorDag;
    LEDs[122] = colorDag;
    LEDs[123] = colorDag;
    LEDs[124] = colorDag;
    LEDs[125] = colorDag;
    LEDs[126] = colorDag;
    LEDs[127] = colorDag;
  } else if (hours >= 12 && hours <= 17) {
    // middag
    LEDs[53] = colorDag;
    LEDs[54] = colorDag;
    LEDs[55] = colorDag;
    LEDs[112] = colorDag;
    LEDs[113] = colorDag;
    LEDs[114] = colorDag;
  } else if (hours >= 18 && hours <= 23) {
    // avond
    LEDs[115] = colorDag;
    LEDs[116] = colorDag;
    LEDs[117] = colorDag;
    LEDs[118] = colorDag;
    LEDs[119] = colorDag;
  }
  // uren
  if (hours == 0) hours = 12;
  if (minutes >= 20 && minutes <= 59) hours++;
  if (hours > 12) hours = hours - 12;
  switch (hours) {
    case 1:  //een
      if (versie == 3) {
        LEDs[81] = colorHour;
        LEDs[82] = colorHour;
        LEDs[83] = colorHour;
      } else {
        LEDs[23] = colorHour;
        LEDs[80] = colorHour;
        LEDs[81] = colorHour;
      }
      break;
    case 2:   // twee
      LEDs[21] = colorHour;
      LEDs[22] = colorHour;
      LEDs[23] = colorHour;
      LEDs[80] = colorHour;
      break;
    case 3:
      if (versie == 3) {
        LEDs[84] = colorHour;
        LEDs[85] = colorHour;
        LEDs[86] = colorHour;
        LEDs[87] = colorHour;
      } else {
        LEDs[83] = colorHour;
        LEDs[84] = colorHour;
        LEDs[85] = colorHour;
        LEDs[86] = colorHour;
      }
      break;
    case 4:
      LEDs[24] = colorHour;
      LEDs[25] = colorHour;
      LEDs[26] = colorHour;
      LEDs[27] = colorHour;
      break;
    case 5:
      LEDs[28] = colorHour;
      LEDs[29] = colorHour;
      LEDs[30] = colorHour;
      LEDs[31] = colorHour;
      break;
    case 6:
      LEDs[88] = colorHour;
      LEDs[89] = colorHour;
      LEDs[90] = colorHour;
      break;
    case 7:
      LEDs[91] = colorHour;
      LEDs[92] = colorHour;
      LEDs[93] = colorHour;
      LEDs[94] = colorHour;
      LEDs[95] = colorHour;
      break;
    case 8:
      LEDs[32] = colorHour;
      LEDs[33] = colorHour;
      LEDs[34] = colorHour;
      LEDs[35] = colorHour;
      break;
    case 9:
      LEDs[36] = colorHour;
      LEDs[37] = colorHour;
      LEDs[38] = colorHour;
      LEDs[39] = colorHour;
      LEDs[96] = colorHour;
      break;
    case 10:
      LEDs[97] = colorHour;
      LEDs[98] = colorHour;
      LEDs[99] = colorHour;
      LEDs[100] = colorHour;
      break;
    case 11:
      LEDs[101] = colorHour;
      LEDs[102] = colorHour;
      LEDs[103] = colorHour;
      break;
    case 12:
      LEDs[40] = colorHour;
      LEDs[41] = colorHour;
      LEDs[42] = colorHour;
      LEDs[43] = colorHour;
      LEDs[44] = colorHour;
      LEDs[45] = colorHour;
      break;
    default:
      // nothing
      break;
  }
  // in de
  LEDs[107] = colorDag;
  LEDs[108] = colorDag;
  LEDs[110] = colorDag;
  LEDs[111] = colorDag;

}

void fillvijf () {
  LEDs[7] = colorMinute;
  LEDs[64] = colorMinute;
  LEDs[65] = colorMinute;
  LEDs[66] = colorMinute;
}

void fillover() {
  LEDs[76] = colorMinute;
  LEDs[77] = colorMinute;
  LEDs[78] = colorMinute;
  LEDs[79] = colorMinute;
}

void filltien () {
  LEDs[68] = colorMinute;
  LEDs[69] = colorMinute;
  LEDs[70] = colorMinute;
  LEDs[71] = colorMinute;
}

void fillkwart () {
  LEDs[8] = colorMinute;
  LEDs[9] = colorMinute;
  LEDs[10] = colorMinute;
  LEDs[11] = colorMinute;
  LEDs[12] = colorMinute;
}

void fillvoor () {
  LEDs[15] = colorMinute;
  LEDs[72] = colorMinute;
  LEDs[73] = colorMinute;
  LEDs[74] = colorMinute;
}

void fillhalf () {
  LEDs[16] = colorHour;
  LEDs[17] = colorHour;
  LEDs[18] = colorHour;
  LEDs[19] = colorHour;
}

void fillminutedetail (byte minutedetail) {
  if (minutedetail > 0) {
    for (int i = 0; i < minutedetail; i++) {
      LEDs[56 + i] = colorMinute;
    }
  }
}

void setColorscheme () {
  switch ( selectedColorscheme ) {
    // Simply add schemes, make sure colorschemeCount increases accordingly
    //  http://fastled.io/docs/3.1/group___colorpalletes.html

    // Palettes
    //
    // RGB Palettes map an 8-bit value (0..255) to an RGB color.
    //
    // You can create any color palette you wish; a couple of starters
    // are provided: Forest, Clouds, Lava, Ocean, Rainbow, and Rainbow Stripes.
    //
    // Palettes come in the traditional 256-entry variety, which take
    // up 768 bytes of RAM, and lightweight 16-entry varieties.  The 16-entry
    // variety automatically interpolates between its entries to produce
    // a full 256-element color map, but at a cost of only 48 bytes or RAM.
    //
    // Basic operation is like this: (example shows the 16-entry variety)
    // 1. Declare your palette storage:
    //    CRGBPalette16 myPalette;
    //
    // 2. Fill myPalette with your own 16 colors, or with a preset color scheme.
    //    You can specify your 16 colors a variety of ways:
    //      CRGBPalette16 myPalette(
    //          CRGB::Black,
    //          CRGB::Black,
    //          CRGB::Red,
    //          CRGB::Yellow,
    //          CRGB::Green,
    //          CRGB::Blue,
    //          CRGB::Purple,
    //          CRGB::Black,
    //
    //          0x100000,
    //          0x200000,
    //          0x400000,
    //          0x800000,
    //
    //          CHSV( 30,255,255),
    //          CHSV( 50,255,255),
    //          CHSV( 70,255,255),
    //          CHSV( 90,255,255)
    //      );
    //
    //    Or you can initiaize your palette with a preset color scheme:
    //      myPalette = RainbowStripesColors_p;
    //
    // 3. Any time you want to set a pixel to a color from your palette, use
    //    "ColorFromPalette(...)" as shown:
    //
    //      uint8_t index = /* any value 0..255 */;
    //      leds[i] = ColorFromPalette( myPalette, index);
    //
    //    Even though your palette has only 16 explicily defined entries, you
    //    can use an 'index' from 0..255.  The 16 explicit palette entries will
    //    be spread evenly across the 0..255 range, and the intermedate values
    //    will be RGB-interpolated between adjacent explicit entries.
    //
    //    It's easier to use than it sounds.
    //

    case 0: currentScheme = CRGBPalette16( CRGB::FloralWhite,
                                             CRGB::Green,
                                             CRGB::Yellow,
                                             CRGB::Purple   ); break;
    case 1: currentScheme = CRGBPalette16( CRGB( 224,  16,   0 ),
                                             CRGB( 192,  64,   0 ),
                                             CRGB( 128, 128,   0 ),
                                             CRGB( 224,  32,   0 ) ); break;
    case 2: currentScheme = CRGBPalette16( CRGB::Aquamarine,
                                             CRGB::Turquoise,
                                             CRGB::Blue,
                                             CRGB::DeepSkyBlue   ); break;
    case 3: currentScheme = RainbowColors_p; break;
    case 4: currentScheme = PartyColors_p; break;
    case 5: currentScheme = CRGBPalette16( CRGB::White ); break;
    case 6: currentScheme = CRGBPalette16( CRGB::LawnGreen ); break;
    case 7: currentScheme = CRGBPalette16( CRGB( 224,   0,  32 ),
                                             CRGB(   0,   0, 244 ),
                                             CRGB( 128,   0, 128 ),
                                             CRGB( 224,   0,  64 ) ); break;
    case 8: currentScheme = ForestColors_p; break;
    case 9: currentScheme = CloudColors_p; break;
    case 10: currentScheme = LavaColors_p; break;
    case 11: currentScheme = OceanColors_p; break;
    case 12: currentScheme = RainbowStripeColors_p; break;

  }
}

void startWiFi() {
  wifiMulti.addAP(ssid, pass);

  Serial.println("Connecting");
  byte i = 0;
  LEDs[i++] = CRGB::Green;
  FastLED.show();
  while (wifiMulti.run() != WL_CONNECTED) {
    delay(250);
    Serial.print('.');
    LEDs[i++] = CRGB::Green;
    FastLED.show();
  }
  Serial.println("\r\n");
  Serial.print("Connected to ");
  Serial.println(WiFi.SSID());
  Serial.print("IP address:\t");
  Serial.print(WiFi.localIP());
  Serial.println("\r\n");
}

void startUDP() {
  Serial.println("Starting UDP");
  UDP.begin(123);                          // Start listening for UDP messages on port 123
  Serial.print("Local port:\t");
  Serial.println(UDP.localPort());
  Serial.println();
}

uint32_t getTime() {
  if (UDP.parsePacket() == 0) { // If there's no response (yet)
    return 0;
  }
  UDP.read(NTPBuffer, NTP_PACKET_SIZE); // read the packet into the buffer
  // Combine the 4 timestamp bytes into one 32-bit number
  uint32_t NTPTime = (NTPBuffer[40] << 24) | (NTPBuffer[41] << 16) | (NTPBuffer[42] << 8) | NTPBuffer[43];
  // Convert NTP time to a UNIX timestamp:
  // Unix time starts on Jan 1 1970. That's 2208988800 seconds in NTP time:
  const uint32_t seventyYears = 2208988800UL;
  // subtract seventy years:
  uint32_t UNIXTime = NTPTime - seventyYears;
  return UNIXTime;
}

void sendNTPpacket(IPAddress& address) {
  memset(NTPBuffer, 0, NTP_PACKET_SIZE);  // set all bytes in the buffer to 0
  // Initialize values needed to form NTP request
  NTPBuffer[0] = 0b11100011;   // LI, Version, Mode
  // send a packet requesting a timestamp:
  UDP.beginPacket(address, 123); // NTP requests are to port 123
  UDP.write(NTPBuffer, NTP_PACKET_SIZE);
  UDP.endPacket();
}

void convertTime(uint32_t time) {
  // Correct time zone
  time += (3600 * timeZone);

  currentDateTime.second = time % 60;
  currentDateTime.minute = time / 60 % 60;
  currentDateTime.hour   = time / 3600 % 24;
  time  /= 60;  // To minutes
  time  /= 60;  // To hours
  time  /= 24;  // To days
  currentDateTime.dayofweek = ((time + 4) % 7) + 1;
  int year = 0;
  int days = 0;
  while ((unsigned)(days += (LEAP_YEAR(year) ? 366 : 365)) <= time) {
    year++;
  }
  days -= LEAP_YEAR(year) ? 366 : 365;
  time  -= days; // To days in this year, starting at 0
  days = 0;
  byte month = 0;
  byte monthLength = 0;
  for (month = 0; month < 12; month++) {
    if (month == 1) { // February
      if (LEAP_YEAR(year)) {
        monthLength = 29;
      } else {
        monthLength = 28;
      }
    } else {
      monthLength = monthDays[month];
    }

    if (time >= monthLength) {
      time -= monthLength;
    } else {
      break;
    }
  }

  currentDateTime.day = time + 1;
  currentDateTime.year = year + 1970;
  currentDateTime.month = month + 1;

  // Correct European Summer time
  if (summerTime()) {
    currentDateTime.hour += 1;
  }

#ifdef DEBUG_ON
  Serial.print(currentDateTime.year);
  Serial.print(" ");
  Serial.print(currentDateTime.month);
  Serial.print(" ");
  Serial.print(currentDateTime.day);
  Serial.print(" ");
  Serial.print(currentDateTime.hour);
  Serial.print(" ");
  Serial.print(currentDateTime.minute);
  Serial.print(" ");
  Serial.print(currentDateTime.second);
  Serial.print(" day of week: ");
  Serial.print(currentDateTime.dayofweek);
  Serial.print(" summer time: ");
  Serial.print(summerTime());
  Serial.println();
#endif
}

boolean summerTime() {

  if (currentDateTime.month < 3 || currentDateTime.month > 10) return false;  // No summer time in Jan, Feb, Nov, Dec
  if (currentDateTime.month > 3 && currentDateTime.month < 10) return true;   // Summer time in Apr, May, Jun, Jul, Aug, Sep
  if (currentDateTime.month == 3 && (currentDateTime.hour + 24 * currentDateTime.day) >= (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 4) % 7)) || currentDateTime.month == 10 && (currentDateTime.hour + 24 * currentDateTime.day) < (3 +  24 * (31 - (5 * currentDateTime.year / 4 + 1) % 7)))
    return true;
  else
    return false;
}
